/**********************************************************************/
/*  SimCell: Cell/B.E. Processor Simulator       Arch Lab. TOKYO TECH */
/**********************************************************************/
#include "define.h"

/**********************************************************************/
enum {
    CH00_SPU_RDEVENTSTAT____ = 0,
    CH01_SPU_WREVENTMASK____ = 1,
    CH02_SPU_WREVENTACK_____ = 2,
    CH03_SPU_RDSIGNOTIFY1___ = 3,
    CH04_SPU_RDSIGNOTIFY2___ = 4,
    CH05_RESERVED___________ = 5,
    CH06_RESERVED___________ = 6,
    CH07_SPU_WRDEC__________ = 7,
    CH08_SPU_RDDEC__________ = 8,
    CH09_SPU_WRMSSYNCREQ____ = 9,
    CH10_RESERVED___________ = 10,
    CH11_SPU_RDEVENTMASK____ = 11,
    CH12_SPU_RDTAGMASK______ = 12,
    CH13_SPU_RDMACHSTAT_____ = 13,
    CH14_SPU_WRSRR0_________ = 14,
    CH15_SPU_RDSRR0_________ = 15,
    CH16_MFC_LSA____________ = 16,
    CH17_MFC_EAH____________ = 17,
    CH18_MFC_EAL____________ = 18,
    CH19_MFC_SIZE___________ = 19,
    CH20_MFC_TAGID__________ = 20,
    CH21_MFC_CMD____________ = 21,
    CH22_MFC_WRTAGMASK______ = 22,
    CH23_MFC_WRTAGUPDATE____ = 23,
    CH24_MFC_RDTAGSTAT______ = 24,
    CH25_MFC_RDLISTSTALLSTAT = 25,
    CH26_MFC_WRLISTSTALLACK_ = 26,
    CH27_MFC_RDATOMICSTAT___ = 27,
    CH28_SPU_WROUTMBOX______ = 28,
    CH29_SPU_RDINMBOX_______ = 29,
    CH30_SPU_WROUTINTRMBOX__ = 30,

    CH_ENTRY01 = 1,
    CH_ENTRY04 = 4,
    CH_ENTRY16 = 16,
    CH_RD_ONLY = 0,
    CH_WR_ONLY = 1,
    CH_NONBLOCK = 0,
    CH_BLOCK = 1,

    DMA_GET = 0x40,
    DMA_PUT = 0x20,
    DMA_SIZE_4BYTE = 4,
    DMA_PACKET_SIZE = 128,
    DMA_SIZE_MAX = 16384,

    MMU_UNITID_UNDEF = -1,
    MMU_UNITID_MEM = 0,

    MFC_CMD_NONE = 0,
    MFC_CMD_EXIST = 1,

    DMA_NONE = -1,
    DMA_PEND = 0,
    DMA_WAIT = 1,
    DMA_RUN = 2,
    DMA_FINISH = 3,

    TAG_UPDATE_IMM = 0,
    TAG_UPDATE_ANY = 1,
    TAG_UPDATE_ALL = 2,
    TAG_UPDATE_UNDEF = 4,

    EIB_CMD_NOTACTIVE = 0,
    EIB_CMD_ACTIVE = 1,

    ENQ_SUCCESS = 0,
    ENQ_FAIL = 1,
    DEQ_SUCCESS = 0,
    DEQ_FAIL = 1,
};

/**********************************************************************/
Channel::Channel(int size, int mode, int type) :
    size(size), blocking_mode(mode), channel_type(type)
{
    message = new uint_032t[size];
    head = 0;
    tail = 0;
    ch_count = (uint_032t) size;
}

/**********************************************************************/
Channel::~Channel()
{
    delete[] message;
}

/**********************************************************************/
inline bool Channel::empty()
{
    return (ch_count == size);
}

/**********************************************************************/
inline bool Channel::full()
{
    return (ch_count == 0);
}

/**********************************************************************/
inline uint_032t Channel::capacity()
{
    return ch_count;
}

/**********************************************************************/
inline uint_032t Channel::occupancy()
{
    return (size - ch_count);
}

/**********************************************************************/
uint_032t Channel::read()
{
    uint_032t data = message[head];
    if (!empty()) {
        head = (head + 1) % size;
        ch_count++;
    }
    return data;
}

/**********************************************************************/
void Channel::write(uint_032t data)
{
    message[tail] = data;
    if (!full()) {
        tail = (tail + 1) % size;
        ch_count--;
    }
}

/**********************************************************************/
ChannelInterface::ChannelInterface()
{
    ch[CH00_SPU_RDEVENTSTAT____] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_RD_ONLY);
    ch[CH01_SPU_WREVENTMASK____] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH02_SPU_WREVENTACK_____] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH03_SPU_RDSIGNOTIFY1___] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_RD_ONLY);
    ch[CH04_SPU_RDSIGNOTIFY2___] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_RD_ONLY);
    ch[CH05_RESERVED___________] = NULL;
    ch[CH06_RESERVED___________] = NULL;
    ch[CH07_SPU_WRDEC__________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH08_SPU_RDDEC__________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_RD_ONLY);
    ch[CH09_SPU_WRMSSYNCREQ____] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_WR_ONLY);
    ch[CH10_RESERVED___________] = NULL;
    ch[CH11_SPU_RDEVENTMASK____] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_RD_ONLY);
    ch[CH12_SPU_RDTAGMASK______] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_RD_ONLY);
    ch[CH13_SPU_RDMACHSTAT_____] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_RD_ONLY);
    ch[CH14_SPU_WRSRR0_________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH15_SPU_RDSRR0_________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_RD_ONLY);
    ch[CH16_MFC_LSA____________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH17_MFC_EAH____________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH18_MFC_EAL____________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH19_MFC_SIZE___________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH20_MFC_TAGID__________] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH21_MFC_CMD____________] =
        new Channel(CH_ENTRY16, CH_BLOCK, CH_WR_ONLY);
    ch[CH22_MFC_WRTAGMASK______] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH23_MFC_WRTAGUPDATE____] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_WR_ONLY);
    ch[CH24_MFC_RDTAGSTAT______] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_RD_ONLY);
    ch[CH25_MFC_RDLISTSTALLSTAT] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_RD_ONLY);
    ch[CH26_MFC_WRLISTSTALLACK_] =
        new Channel(CH_ENTRY01, CH_NONBLOCK, CH_WR_ONLY);
    ch[CH27_MFC_RDATOMICSTAT___] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_RD_ONLY);
    ch[CH28_SPU_WROUTMBOX______] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_WR_ONLY);
    ch[CH29_SPU_RDINMBOX_______] =
        new Channel(CH_ENTRY04, CH_BLOCK, CH_RD_ONLY);
    ch[CH30_SPU_WROUTINTRMBOX__] =
        new Channel(CH_ENTRY01, CH_BLOCK, CH_WR_ONLY);
}

/**********************************************************************/
ChannelInterface::~ChannelInterface()
{
    for (int i = 0; i < CH_NUM; i++) {
        delete ch[i];
        ch[i] = NULL;
    }
}

/**********************************************************************/
bool ChannelInterface::spureadable(int num)
{
    if (ch[num] == NULL)
        return 0;

    return !(ch[num]->channel_type == CH_WR_ONLY ||
             (ch[num]->blocking_mode == CH_BLOCK && ch[num]->empty()));
}

/**********************************************************************/
bool ChannelInterface::spuwritable(int num)
{
    if (ch[num] == NULL)
        return 0;

    return !(ch[num]->channel_type == CH_RD_ONLY ||
             (ch[num]->blocking_mode == CH_BLOCK && ch[num]->full()));
}

/**********************************************************************/
int ChannelInterface::spureadch(int num, uint_032t * data)
{
    if (!spureadable(num))
        return 1;

    *data = ch[num]->read();
    return 0;
}

/**********************************************************************/
int ChannelInterface::spuwritech(int num, uint_032t data)
{
    if (!spuwritable(num))
        return 1;

    ch[num]->write(data);
    return 0;
}

/**********************************************************************/
uint_032t ChannelInterface::spureadchcnt(int num)
{
    if (ch[num] == NULL)
        return 0;
    if (ch[num]->blocking_mode == CH_NONBLOCK)
        return 1;
    if (ch[num]->channel_type == CH_RD_ONLY)
        return ch[num]->occupancy();
    if (ch[num]->channel_type == CH_WR_ONLY)
        return ch[num]->capacity();
    return 0;
}

/**********************************************************************/
void ChannelInterface::mfcreadch(int num, uint_032t * data)
{
    *data = ch[num]->read();
}

/**********************************************************************/
void ChannelInterface::mfcwritech(int num, uint_032t data)
{
    ch[num]->write(data);
}

/**********************************************************************/
Mfc::Mfc(Chip * chip, Spe * spe)
{
    this->chip = chip;
    this->mmu = chip->mmu;
    this->spe = spe;
    this->chi = spe->chi;

    head = 0;
    tail = 0;
    entry = 0;
    for (int i = 0; i < MFC_MAX_ENTRY; i++)
        cmdtag[i].state = DMA_NONE;
    tagmask = 0;
    tagupdate = TAG_UPDATE_UNDEF;
    tagstat = 0;
    decrementer = 0xffffffff;
}

/**********************************************************************/
inline dmacmd_t Mfc::getcmdfromchi()
{
    dmacmd_t cmd;
    uint_032t classid_cmd;

    chi->mfcreadch(CH16_MFC_LSA____________, &cmd.lsa.physical);
    chi->mfcreadch(CH18_MFC_EAL____________, &cmd.ea.logical);
    chi->mfcreadch(CH19_MFC_SIZE___________, &cmd.size);
    chi->mfcreadch(CH20_MFC_TAGID__________, &cmd.tagid);
    chi->mfcreadch(CH21_MFC_CMD____________, &classid_cmd);

    cmd.classid = classid_cmd >> 16;
    cmd.cmd = classid_cmd & 0xffff;
    cmd.lsa.unitid = spe->unitid;
    return cmd;
}

/**********************************************************************/
inline int Mfc::enq_spucmd(dmacmd_t cmd, dmacmdtag_t tag)
{
    if (entry == MFC_MAX_ENTRY)
        return ENQ_FAIL;
    else {
        mfcspucmdq[tail] = cmd;
        cmdtag[tail] = tag;
        entry++;
        tail = (tail + 1) % MFC_MAX_ENTRY;
        return ENQ_SUCCESS;
    }
}

/**********************************************************************/
inline int Mfc::deq_spucmd()
{
    if (entry == 0)
        return DEQ_FAIL;
    else {
        cmdtag[head].state = DMA_NONE;
        entry--;
        head = (head + 1) % MFC_MAX_ENTRY;
        return DEQ_SUCCESS;
    }
}

/**********************************************************************/
inline bool Mfc::ismemread(dmacmd_t cmd)
{
    return ((cmd.ea.unitid == MMU_UNITID_MEM)
            && (cmd.cmd == DMA_GET));
}

/**********************************************************************/
inline bool Mfc::ismemwrite(dmacmd_t cmd)
{
    return ((cmd.ea.unitid == MMU_UNITID_MEM)
            && (cmd.cmd == DMA_PUT));
}

/**********************************************************************/
inline int Mfc::getdelay(dmacmd_t cmd)
{
    int packet_num = (cmd.size / DMA_PACKET_SIZE);
    if (ismemread(cmd))
        return chip->mem_rddelay + packet_num * chip->mem_width;
    else if (ismemwrite(cmd))
        return chip->mem_wrdelay + packet_num * chip->mem_width;
    else
        return chip->c2c_delay + packet_num * chip->c2c_width;
}

/**********************************************************************/
inline void Mfc::makedmacmd()
{
    dmacmd_t cmd = getcmdfromchi();
    mmu->transaddr(&cmd.ea);
    dmacmdtag_t tag;
    tag.delay = getdelay(cmd);
    tag.state = DMA_PEND;
    
    enq_spucmd(cmd, tag);
    
    if (chip->log_mode)
        for (uint i = 0; i < cmd.size; i += 16)
            chip->setlogdata(spe->unitid, cmd.lsa.physical + i, MEM_DMA);
}

/**********************************************************************/
inline uint_032t Mfc::maketagstat()
{
    uint_032t stat = tagmask;
    for (int i = 0; i < entry; i++) {
        int index = (head + i) % MFC_MAX_ENTRY;
        switch (cmdtag[index].state) {
        case DMA_PEND:
        case DMA_WAIT:
        case DMA_RUN:
            stat &= ~(1 << mfcspucmdq[index].tagid);
            break;
        case DMA_FINISH:
        case DMA_NONE:
            stat |= (1 << mfcspucmdq[index].tagid);
            break;
        default:
            printf("## Mfc::maketagstat(), undefined TagStat.\n");
            exit(1);
        }
    }
    return (stat & tagmask);
}

/**********************************************************************/
inline void Mfc::checkchannel()
{
    if (!chi->ch[CH07_SPU_WRDEC__________]->empty()) {
        chi->mfcreadch(CH07_SPU_WRDEC__________, &decrementer);
        chi->mfcwritech(CH08_SPU_RDDEC__________, decrementer);
    }
    if (!chi->ch[CH21_MFC_CMD____________]->empty() &&
        entry < MFC_MAX_ENTRY)
        makedmacmd();
    if (!chi->ch[CH22_MFC_WRTAGMASK______]->empty())
        chi->mfcreadch(CH22_MFC_WRTAGMASK______, &tagmask);
    if (!chi->ch[CH23_MFC_WRTAGUPDATE____]->empty()) {
        chi->mfcreadch(CH23_MFC_WRTAGUPDATE____, &tagupdate);
    }
}

/**********************************************************************/
inline void Mfc::checkdma()
{
    for (int i = 0, deqflag = 0; i < entry; i++) {
        int index = (head + i) % MFC_MAX_ENTRY;
        switch (cmdtag[index].state) {
        case DMA_PEND:
            if (chip->eib->enq_waitcmd(&mfcspucmdq[index], &cmdtag[index])
                == ENQ_SUCCESS)
                cmdtag[index].state = DMA_WAIT;
            break;
        case DMA_WAIT:
            break;
        case DMA_RUN:
            cmdtag[index].delay--;
            break;
        case DMA_FINISH:
            cmdtag[index].state = DMA_NONE;
            break;
        case DMA_NONE:
            if (i == deqflag) {
                deq_spucmd();
                deqflag++;
            }
            break;
        default:
            printf("## Mfc::checkdma(), undefined TagStat.\n");
            exit(1);
        }
    }
}

/**********************************************************************/
inline void Mfc::updatetagstate()
{
    switch (tagupdate) {
    case TAG_UPDATE_UNDEF:
        break;;
    case TAG_UPDATE_IMM:
        tagstat = maketagstat();
        chi->mfcwritech(CH24_MFC_RDTAGSTAT______, tagstat);
        tagupdate = TAG_UPDATE_UNDEF;
        break;
    case TAG_UPDATE_ANY:
        tagstat = maketagstat();
        if (0 < tagstat && tagstat <= tagmask) {
            chi->mfcwritech(CH24_MFC_RDTAGSTAT______, tagstat);
            tagupdate = TAG_UPDATE_UNDEF;
        }
        break;
    case TAG_UPDATE_ALL:
        tagstat = maketagstat();
        if (tagstat == tagmask) {
            chi->mfcwritech(CH24_MFC_RDTAGSTAT______, tagstat);
            tagupdate = TAG_UPDATE_UNDEF;
        }
        break;
    default:
        printf("## Mfc::updatetagstate(), illegal request.\n");
        exit(1);
    }
}

/**********************************************************************/
inline void Mfc::decrementerstep()
{
    if (chip->cycle % 40 == 0) {
        decrementer--;
        chi->mfcwritech(CH08_SPU_RDDEC__________, decrementer);
    }
}

/**********************************************************************/
void Mfc::step()
{
    checkchannel();
    checkdma();
    updatetagstate();
    decrementerstep();
}

/**********************************************************************/
Mmu::Mmu(Chip * chip)
{
    chip->mmu = this;
    for (int i = 0; i < BLOCK_TABLE_SIZE; i++) {
        mmutable[i].addr = i;
        mmutable[i].id = MMU_UNITID_MEM;
    }
}

/**********************************************************************/
void Mmu::transaddr(address_t * addr)
{
    uint_032t block = addr->logical / BLOCK_SIZE;
    uint_032t offset = addr->logical % BLOCK_SIZE;

    addr->physical = (mmutable[block].addr * BLOCK_SIZE) + offset;
    addr->unitid = mmutable[block].id;
}

/**********************************************************************/
int Mmu::mapping(uint_032t ea, int id, uint_032t addr)
{
    uint_032t block = ea / BLOCK_SIZE;
    uint_032t block_addr = addr / BLOCK_SIZE;

    mmutable[block].addr = block_addr;
    mmutable[block].id = id;
    return 0;
}

/**********************************************************************/
void Mmu::mapls(uint_032t ea, int id)
{
    int i;
    for (i = 0; i < LS_SIZE; i += BLOCK_SIZE)
        mapping(ea + i, id, i);
}

/**********************************************************************/
Eib::Eib(Chip * chip)
{
    chip->eib = this;
    this->chip = chip;
    this->mem = chip->mem;
    for (int i = 0; i < MAX_SPES; i++) {
        if (i < chip->spe_num)
            this->spe[i] = chip->spe[i];
        else
            this->spe[i] = NULL;
    }
    for (int i = 0; i < EIB_MAX_ENTRY; i++)
        valid[i] = EIB_CMD_NOTACTIVE;
    entry_count = 0;
    head = 0;
    tail = 0;
    transdata_ptr = new uint_032t[DMA_SIZE_MAX / sizeof(uint_032t)];
}

/**********************************************************************/
Eib::~Eib()
{
    delete [] transdata_ptr;
    transdata_ptr = NULL;
}

/**********************************************************************/
inline bool Eib::emptyactivecmd()
{
    return (entry_count == 0);
}

/**********************************************************************/
inline bool Eib::fullactivecmd()
{
    return (entry_count == EIB_MAX_ENTRY);
}

/**********************************************************************/
inline int Eib::enq_waitcmd(dmacmd_t * cmd, dmacmdtag_t * tag)
{
    if (head == (tail + 1) % EIB_WAIT_ENTRY)
        return ENQ_FAIL;
    else {
        waitcmdq[tail].cmd = cmd;
        waitcmdq[tail].tag = tag;
        tail = (tail + 1) % EIB_WAIT_ENTRY;
        return ENQ_SUCCESS;
    }
}

/**********************************************************************/
inline int Eib::deq_waitcmd(eibcmd_t * eibcmd)
{
    if (head == tail)
        return DEQ_FAIL;
    else {
        *eibcmd = waitcmdq[head];
        head = (head + 1) % EIB_WAIT_ENTRY;
        return DEQ_SUCCESS;
    }
}

/**********************************************************************/
inline void Eib::checkcmd(dmacmd_t *cmd)
{
    int dma_vio = 0;
    static int warn = 0;

    if ((cmd->cmd != DMA_GET) && (cmd->cmd != DMA_PUT)) {
        printf("## undefined DMA Command Error\n");
        exit(1);
    }

    if ((cmd->size == 0) || (cmd->size > DMA_SIZE_MAX)) {
        dma_vio = 1;
    } else if (cmd->size < 16) {
        dma_vio = ((16 % cmd->size != 0) ||
                   (cmd->ea.physical % 16 != cmd->lsa.physical % 16) ||
                   (cmd->ea.physical % cmd->size != 0) ||
                   (cmd->lsa.physical % cmd->size != 0));
    } else {
        dma_vio = ((cmd->size % 16 != 0) ||
                   (cmd->ea.physical % 16 != 0) ||
                   (cmd->lsa.physical % 16 != 0));
    }
    if (dma_vio) {
        if (!warn) {
            fprintf(stderr, "## DMA Access Violation Detected.\n");
            fprintf(stderr, "## EA=%08x LS=%08x SIZE=%5d\n",
                    (uint) cmd->ea.physical, (uint) cmd->lsa.physical,
                    (uint) cmd->size);
            warn = 1;
            if (!chip->force_mode) {
                fprintf(stderr, "## Simulaton aborted.\n");
                exit(1);
            }
        }
    }
}

/**********************************************************************/
inline void Eib::transdata(dmacmd_t * cmd)
{
    int src_id, dst_id, size = cmd->size;
    uint_032t src_addr, dst_addr;

    checkcmd(cmd);

    if (cmd->cmd == DMA_GET) {
        src_addr = cmd->ea.physical;
        dst_addr = cmd->lsa.physical;
        src_id = cmd->ea.unitid;
        dst_id = cmd->lsa.unitid;
    } else { // if (cmd->cmd == DMA_PUT) {
        src_addr = cmd->lsa.physical;
        dst_addr = cmd->ea.physical;
        src_id = cmd->lsa.unitid;
        dst_id = cmd->ea.unitid;
    }

    if (src_id == MMU_UNITID_MEM) {
        mem->readnb(src_addr, size, transdata_ptr);
        spe[dst_id - 1]->ls->writenb(dst_addr, size, transdata_ptr);
    } else if (dst_id == MMU_UNITID_MEM) {
        spe[src_id - 1]->ls->readnb(src_addr, size, transdata_ptr);
        mem->writenb(dst_addr, size, transdata_ptr);
    } else {
        spe[src_id - 1]->ls->readnb(src_addr, size, transdata_ptr);
        spe[dst_id - 1]->ls->writenb(dst_addr, size, transdata_ptr);
    }
}

/**********************************************************************/
inline void Eib::comstep()
{
    for (int i = 0; i < EIB_MAX_ENTRY; i++) {
        if (emptyactivecmd())
            return;
        if (valid[i] == EIB_CMD_NOTACTIVE)
            continue;
        if (activecmd[i].tag->delay <= 0) {
            transdata(activecmd[i].cmd);
            activecmd[i].tag->state = DMA_FINISH;
            valid[i] = EIB_CMD_NOTACTIVE;
            entry_count--;
        }
    }
}

/**********************************************************************/
inline void Eib::activatecmd()
{
    for (int i = 0; i < EIB_MAX_ENTRY; i++) {
        if (fullactivecmd())
            return;
        if (valid[i] == EIB_CMD_NOTACTIVE) {
            if (deq_waitcmd(&activecmd[i]) == DEQ_SUCCESS) {
                activecmd[i].tag->state = DMA_RUN;
                valid[i] = EIB_CMD_ACTIVE;
                entry_count++;
            } else
                return;
        }
    }
}

/**********************************************************************/
void Eib::step()
{
    activatecmd();
    comstep();
}

/**********************************************************************/
