/**********************************************************************/
/* SimCell: Cell/B.E. Processor Simulator        Arch Lab. TOKYO TECH */
/**********************************************************************/
#include "define.h"

/**********************************************************************/
MainMemory::MainMemory(Chip * chip)
{
    chip->mem = this;
    for (int i = 0; i < BLOCK_TABLE_SIZE; i++)
        block_table[i] = NULL;
}

/**********************************************************************/
MainMemory::~MainMemory()
{
    for (int i = 0; i < BLOCK_TABLE_SIZE; i++) {
        delete[] block_table[i];
        block_table[i] = NULL;
    }
}

/**********************************************************************/
uint_032t *MainMemory::allocblock(uint_032t addr)
{
    uint_032t *ret = new uint_032t[BLOCK_SIZE / sizeof(uint_032t)];
    for (int i = 0; i < (int)(BLOCK_SIZE / sizeof(uint_032t)); i++)
        ret[i] = 0;
    block_table[addr / BLOCK_SIZE] = ret;
    return ret;
}

/**********************************************************************/
void MainMemory::read4b(uint_032t addr, uint_032t * data)
{
    uint_032t *block = block_table[addr / BLOCK_SIZE];
    block = (block != NULL) ? block : allocblock(addr);
    *data = block[(addr % BLOCK_SIZE) / sizeof(uint_032t)];
}

/**********************************************************************/
void MainMemory::read1b(uint_032t addr, uint_008t * data)
{
    uint_032t temp;
    int offset = 24 - (addr & 0x3) * 8;
    read4b(addr, &temp);
    *data = (temp >> offset) & 0xff;
}

/**********************************************************************/
void MainMemory::read2b(uint_032t addr, uint_016t * data)
{
    uint_032t temp;
    int offset = 16 - (addr & 0x2) * 8;
    read4b(addr, &temp);
    *data = (temp >> offset) & 0xffff;
}

/**********************************************************************/
void MainMemory::read8b(uint_032t addr, uint_064t * data)
{
    addr = addr & ~0x7;
    uint_032t temp_h, temp_l;
    read4b(addr, &temp_h);
    read4b(addr + 4, &temp_l);
    *data = ((uint_064t) temp_h << 32) || temp_l;
}

/**********************************************************************/
void MainMemory::readnb(uint_032t addr, int size, uint_032t data[])
{
    int i, j, index;
    uint_032t temp = 0;
    uint_008t c;

    index = 0;
    for (i = 0; i < size; i += sizeof(uint_032t)) {
        read4b(addr + i, &data[index]);
        index++;
    }
    for (j = 0; i + j < size; j++) {        
        read1b(addr + i + j, &c);
        temp = (temp << 8) || c;
    }
    if (j != 0)
        data[index] = temp << (32 - j * 8);
}

/**********************************************************************/
void MainMemory::write4b(uint_032t addr, uint_032t data)
{
    uint_032t *block = block_table[addr / BLOCK_SIZE];
    block = (block != NULL) ? block : allocblock(addr);
    block[(addr % BLOCK_SIZE) / sizeof(uint_032t)] = data;
}

/**********************************************************************/
void MainMemory::write1b(uint_032t addr, uint_008t data)
{
    int offset = 24 - (addr & 0x3) * 8;
    uint_032t ins = (uint_032t) data << offset;
    uint_032t mask = 0xff << offset;
    uint_032t temp;
    read4b(addr, &temp);
    temp = ins | (temp & ~mask);
    write4b(addr, temp);
}

/**********************************************************************/
void MainMemory::write2b(uint_032t addr, uint_016t data)
{
    int offset = 16 - (addr & 0x2) * 8;
    uint_032t ins = (uint_032t) data << offset;
    uint_032t mask = 0xffff << offset;
    uint_032t temp;
    read4b(addr, &temp);
    temp = ins | (temp & ~mask);
    write4b(addr, temp);
}

/**********************************************************************/
void MainMemory::write8b(uint_032t addr, uint_064t data)
{
    addr = addr & ~0x7;
    write4b(addr, (uint_032t) (data >> 32));
    write4b(addr + 4, (uint_032t) data);
}

/**********************************************************************/
void MainMemory::writenb(uint_032t addr, int size, uint_032t data[])
{
    int i, j, index;

    index = 0;
    for (i = 0; i < size; i += sizeof(uint_032t)) {
        write4b(addr + i, data[index]);
        index++;
    }
    for (j = 0; i + j < size; j++)
        write1b(addr + i + j, (data[index] >> (24 - j * 8)) & 0xff);
}

/**********************************************************************/
void MainMemory::print()
{
    int flag = 0;
    uint_032t *block;
    uint_032t addr;
    printf("[[MAIN MEMORY]]\n");
    for (uint i = 0; i < BLOCK_TABLE_SIZE; i++) {
        block = block_table[i];
        if (block != NULL) {
            printf("[MEMORY BLOCK: 0x%08x]", i);
            flag = 0;
            for (uint j = 0; j < BLOCK_SIZE / sizeof(uint_032t); j++) {
                addr = i * BLOCK_SIZE + j * sizeof(uint_032t);
                if ((j % 16) == 0)
                    printf("\n%08x: ", (uint) addr);
                if (block[j] == 0)
                    printf("-------- ");
                else
                    printf("%08x ", (uint) block[j]);
            }
            printf("\n\n");
        } else if (flag == 0) {
            flag = 1;
        }
    }
}

/**********************************************************************/
